####
#### TkXtra.py -- Extended widgets for Tkinter
#### Copyright (C) 1996, Sudhir Shenoy.

#### This library is free software; you can redistribute it and/or
#### modify it under the terms of the GNU Library General Public
#### License as published by the Free Software Foundation; either
#### version 2 of the License, or (at your option) any later version.
#### 
#### This library is distributed in the hope that it will be useful,
#### but WITHOUT ANY WARRANTY; without even the implied warranty of
#### MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#### Library General Public License for more details.
#### 
#### You should have received a copy of the GNU Library General Public
#### License along with this library; if not, write to the
#### Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
#### MA 02139, USA.

#### Notes:
####
#### Many of the widgets here are like the ones in Tix (Tk based extended
#### widget set). The decision to develop these widgets was made when
#### Tix decided to go commercial (a decision which has since been reversed).
#### However, I think the other widgets here are useful in their own right.
#### Besides, since this is entirely in Python, we don't have any
#### version synchronization problem (only a working Tkinter is required).
####
#### Unfortunately, most of the widgets here are subclassed from the Frame
#### widget rather than the corresponding main widget. For example,
#### it would have been nice to have subclassed LabelledEntry from the
#### Entry widget rather than the Frame widget. There is a hack which
#### allows this (somewhere in the Python demo directory) but I didn't
#### feel like using it :-) So one is forced to use LabelledEntry.Entry
#### to access the Entry widget properties
####

import string
from types import IntType, StringType, ListType, TupleType

from Tkinter import *
from Tkinter import _cnfmerge

import DateHols				# needed by DateEntry
from StrConv import Str2Amt, Amt2Str	# for NumericEntry

# Set up default date printing format as e.g., 'Fri Mar 23 96' (saves space)
DateHols.LONG_DATE_FORMAT = DateHols.SHORT_WMDY

AUTO = 'auto'


#########################################################################
###                                                                   ###
### Pure Combination widgets (no validation) ---                      ###
###              ScrolledList, PopupList etc.                         ###
###                                                                   ###
#########################################################################

# General scrolled widget. Implements scrollbars and their display policy only.
# Not meant to be instantiated.
# Scrollbar may be specified as auto,x,y,none or both. In case of auto,
# scrollbars are displayed only if needed
class ScrolledWidget:
    def __init__(self, scrollbar=AUTO, widget):
	if scrollbar not in [AUTO, NONE, BOTH, X, Y]:
	    scrollbar = AUTO
	self.sb_policy = scrollbar
	self.widget = widget
	self.widget.bind('<Configure>', self.redisplay)

	self.vsb = Scrollbar(self)
	self.widget['yscrollcommand'] = (self.vsb, 'set')
	self.vsb.config(command=(self.widget, 'yview'))

	self.hsb = Scrollbar(self, orient=HORIZONTAL)
	self.widget['xscrollcommand'] = (self.hsb, 'set')
	self.hsb.config(command=(self.widget, 'xview'))

    def __setitem__(self, key, value):
	if key == 'scrollbar':
	    if value not in [AUTO, NONE, BOTH, X, Y]:
		value = AUTO
	    self.sb_policy = value
	    self.redisplay()
	else:
	    Widget.config(self, {key: value})

    def __getitem__(self, key):
	if key == 'scrollbar':
	    return self.sb_policy
	else:
	    Widget.__getitem__(self, key)

    def check_dimensions(self):
	if not self.widget.winfo_ismapped():
	    return []
	xv = string.split(self.tk.call(self.widget._w, 'xview'))
	yv = string.split(self.tk.call(self.widget._w, 'yview'))
	xmin = string.atof(xv[0])
	xmax = string.atof(xv[1])
	ymin = string.atof(yv[0])
	ymax = string.atof(yv[1])
	retval = []
	if xmin > 0 or xmax < 1:
	    retval.append(X)
	if ymin > 0 or ymax < 1:
	    retval.append(Y)
	return retval

    # Check if scrollbars need to be displayed
    def redisplay(self, event=None):
	sb_display_list = []
	if self.sb_policy == NONE:
	    pass
	elif self.sb_policy in [X, Y, BOTH]:
	    if self.sb_policy != X:
		sb_display_list.append(Y)
	    if self.sb_policy != Y:
		sb_display_list.append(X)
	else:
	    sb_display_list = self.check_dimensions()
	if Y in sb_display_list:
	    self.vsb.pack(side=RIGHT, fill=Y, before=self.widget)
	else:
	    self.vsb.forget()
	if X in sb_display_list:
	    self.hsb.pack(side=BOTTOM, fill=X, before=self.widget)
	else:
	    self.hsb.forget()

# A Text widget with (optionally displayed) scrollbars. 
# The Text is accessible as .text
class ScrolledText(Frame, ScrolledWidget):
    def __init__(self, master=None, scrollbar=AUTO, cnf={}, **kw):
	if kw:
	    cnf = _cnfmerge((cnf, kw))
	else:
	    cnf = _cnfmerge(cnf)
	Frame.__init__(self, master, cnf)

	self.text = Text(self, wrap=NONE)
	self.text.pack(side=LEFT, expand=1, fill=BOTH)

	ScrolledWidget.__init__(self, scrollbar=scrollbar, widget=self.text)

# A listbox with scrollbar. Avoids having to have yview & yscrollcommand
# all over the place.
# The listbox is accessible as .listbox
# A double-click procedure may be assigned for convenience.
class ScrolledList(Frame, ScrolledWidget):
    def __init__(self, master=None, init_vals=None, variable=None,
		 dbl_click=None, scrollbar=AUTO, cnf={}, **kw):
	if kw:
	    cnf = _cnfmerge((cnf, kw))
	else:
	    cnf = _cnfmerge(cnf)
	Frame.__init__(self, master, cnf)
	self.listbox = Listbox(self)
	ScrolledWidget.__init__(self, scrollbar=scrollbar, widget=self.listbox)

	self.listbox.pack(side=RIGHT, expand=1, fill=BOTH)
	self.listbox.bind('<ButtonRelease-1>', self.select_event)
	if dbl_click:
	    self.listbox.bind('<Double-ButtonRelease-1>', self.exec_event)

	if init_vals:
	    for x in init_vals:
		self.listbox.insert(END, x)
	self.variable = variable
	self.dbl_click = dbl_click

    def add(self, string):
	self.listbox.insert(END, string)
	self.redisplay()

    def add_list(self, list):
	for item in list:
	    self.listbox.insert(END, item)
	self.redisplay()

    def remove(self, idx):
	if type(idx) == IntType and idx < self.listbox.size():
	    self.listbox.delete(idx)
	    return

	if type(idx) != StringType:
	    idx = `idx`
	for i in range(self.listbox.size()):
	    if self.listbox.get(i) == idx:
		self.listbox.delete(i)
		return
	# not found
	return

    def remove_all(self):
	self.listbox.delete(0, END)
	self.redisplay()

    def select_event(self, event):
	if self.variable:
	    idx = self.listbox.curselection()[0]
	    s = self.listbox.get(idx)
	    self.variable.set(s)

    def exec_event(self, event):
	if self.dbl_click:
	    idx = self.listbox.curselection()[0]
	    s = self.listbox.get(idx)
	    apply(self.dbl_click, (s,))

# A popup which contains a scrolled listbox. Used by ComboBox for
# the dropdown list. ref_widget is the widget below which the popup
# will be displayed.
class PopupList(ScrolledList):
    def __init__(self, ref_widget=None, init_vals=None, callback=None,
		 cnf={}, **kw):
	if kw:
	    cnf = _cnfmerge((cnf, kw))
	else:
	    cnf = _cnfmerge(cnf)

	self.toplevel = Toplevel()
	self.toplevel['relief'] = RAISED
	self.toplevel['bd'] = 1
	self.toplevel['cursor'] = 'arrow'
	self.toplevel.overrideredirect(1)
	self.toplevel.withdraw()
	self.popped_up = 0
	ScrolledList.__init__(self, master=self.toplevel, init_vals=init_vals,
			      cnf=cnf)
	self.callback = callback
	self.ref_widget = ref_widget
	self.pack(expand=1, fill=BOTH)

    def show_hide(self, event=None):
	if self.popped_up:
	    self.hide(event)
	else:
	    self.show(event)

    def show(self, event=None):
	x = 0
	y = 0
	if self.ref_widget:
	    x = self.ref_widget.winfo_rootx()
	    y = self.ref_widget.winfo_rooty()
	    y = y + self.ref_widget.winfo_height()
	elif event:
	    x = event.x_root
	    y = event.y_root
	scrheight = self.winfo_screenheight()
	scrwidth = self.winfo_screenwidth()
	width = self.listbox.winfo_width()
	height = self.listbox.winfo_height()
	if x + width > scrwidth:
	    x = scrwidth - width
	if y + height > scrheight:
	    y = scrheight - height
	s = '+%d+%d' % (x, y)
	self.toplevel.geometry(s)
	self.toplevel.deiconify()
	self.toplevel.tkraise()
	self.toplevel.focus_set()
	self.popped_up = 1

    def hide(self, event=None):
	self.toplevel.withdraw()
	self.popped_up = 0

    def select_event(self, event):
	if self.callback:
	    idx = self.listbox.curselection()[0]
	    s = self.listbox.get(idx)
	    self.callback(s)
	self.hide()

# ButtonBox - A frame to enclose buttons either horizontally or vertically.
# Setting the state of the box to disabled disables all the buttons.
class ButtonBox(Frame):
    def __init__(self, master=None, state=NORMAL, orient=HORIZONTAL,
		 cnf={}, **kw):
	if kw:
	    cnf = _cnfmerge((cnf, kw))
	else:
	    cnf = _cnfmerge(cnf)
	Frame.__init__(self, master, cnf)
	self.orient = orient
	self.state = state
	self.btns = []

    def __setitem__(self, key, value):
	if key in ['state', 'orient']:
	    if key == 'state':
		self.state = value
		self.able()
	    elif key == 'orient':
		self.orient = value
		self.repaint()
	else:
	    Widget.config(self, {key: value})

    def __getitem__(self, key):
	if key in ['state', 'orient']:
	    if key == 'state':
		return self.state
	    elif key == 'orient':
		return self.orient
	else:
	    Widget.__getitem__(self, key)

    def able(self):
	for b in self.btns:
	    b['state'] = self.state

    def repaint(self):
	for b in self.btns:
	    if self.orient == HORIZONTAL:
		b.pack(side=LEFT)
	    else:
		b.pack(side=TOP)

    def add(self, cnf={}, **kw):
	if kw:
	    cnf = _cnfmerge((cnf, kw))
	else:
	    cnf = _cnfmerge(cnf)
	b = Button(self, cnf)
	if self.orient == HORIZONTAL:
	    b.pack(side=LEFT)
	else:
	    b.pack(side=TOP)
	self.btns.append(b)
	return b

    def enable(self):
	self.state = NORMAL
	self.able()

    def disable(self):
	self.state = DISABLED
	self.able()


# Like a radio button / check button box except that these buttons are
# drawn as regular buttons (look bigger and more uniform). If labelside is
# not specified (as TOP/BOTTOM/LEFT/RIGHT), it is drawn as TOP for vertically
# oriented boxes and LEFT otherwise
class SelectBox(Frame):
    def __init__(self, master=None, label=None, radio=1, allowzero=1,
		 orient=HORIZONTAL, labelside=None, labelwidth=None, anchor=NE,
		 variable=None, state=NORMAL, command=None,
		 cmd_param=None, cnf={}, **kw):
	if kw:
	    cnf = _cnfmerge((cnf, kw))
	else:
	    cnf = _cnfmerge(cnf)
	Frame.__init__(self, master, cnf)
	if label:
	    self.label = Label(self, text=label, anchor=anchor)
	else:
	    self.label = Label(self, text='', anchor=anchor)
	if labelwidth:
	    self.label['width'] = labelwidth
	self.btnbox = Frame(self, relief=SUNKEN, borderwidth=1)
	if orient not in [HORIZONTAL, VERTICAL]:
	    orient = HORIZONTAL
	self.orient = orient
	self.radio = radio
	if self.radio:
	    self.allowzero = 0
	else:
	    self.allowzero = allowzero
	if labelside not in [TOP, BOTTOM, LEFT, RIGHT]:
	    if self.orient == HORIZONTAL:
		labelside = LEFT
	    else:
		labelside = TOP
	self.labelside = labelside
	if state not in [NORMAL, DISABLED]:
	    state = NORMAL
	self.state = state
	if variable:
	    self.variable = variable
	else:
	    self.variable = StringVar()
	self.value = []
	self.user_proc = command
	self.user_param = cmd_param

	self.btns = []
	self.selected_btn = []
	self.label.pack(side=self.labelside)
	self.btnbox.pack(side=self.labelside)

    def __setitem__(self, key, value):
	if key in ['label', 'radio', 'allowzero', 'orient', 'labelside',
		   'labelwidth', 'anchor', 'variable', 'state', 'command',
		   'cmd_param', 'value']:
	    if key == 'label':
		self.label['text'] = value
	    elif key == 'radio':
		pass			# XXXXX
	    elif key == 'allowzero':
		pass			# XXXXX
	    elif key == 'orient':
		pass			# XXXXX
	    elif key == 'labelside':
		pass			# XXXXX
	    elif key == 'variable':
		if value:
		    self.variable = value
		else:
		    self.variable = StringVar()
	    elif key == 'state':
		if value == DISABLED:
		    self.state = DISABLED
		else:
		    self.state = NORMAL
		for b in self.btns:
		    b['state'] = self.state
	    elif key == 'command':
		self.user_proc = value
	    elif key == 'cmd_param':
		self.user_param = value
	    elif key == 'value':
		pass			# XXXXX
	    elif key == 'labelwidth':
		self.label['width'] = value
	    elif key == 'anchor':
		self.label['anchor'] = value
	else:
	    Widget.config(self, {key: value})

    def __getitem__(self, key):
	if key in ['label', 'radio', 'allowzero', 'orient', 'labelside',
		   'labelwidth', 'anchor', 'variable', 'state', 'command',
		   'cmd_param', 'value']:
	    if key == 'label':
		return self.label['text']
	    elif key == 'radio':
		return self.radio
	    elif key == 'allowzero':
		return self.radio
	    elif key == 'orient':
		return self.radio
	    elif key == 'labelside':
		return self.labelside
	    elif key == 'variable':
		return self.variable
	    elif key == 'state':
		return self.state
	    elif key == 'command':
		return self.user_proc
	    elif key == 'cmd_param':
		return self.user_param
	    elif key == 'value':
		return self.value
	    elif key == 'labelwidth':
		return self.label['width']
	    elif key == 'anchor':
		return self.label['anchor']
	else:
	    Widget.__getitem__(self, key)

    def add(self, retval, text=None, image=None):
	if image:
	    b = Button(self.btnbox, image=image)
	else:
	    b = Button(self.btnbox, text=text, highlightthickness=0)
	b['command'] = lambda w=b, s=self: s.btn_press(w)
	b.retval = retval
	if self.state == DISABLED:
	    b['state'] = DISABLED
	if self.orient == HORIZONTAL:
	    b.pack(side=LEFT)
	else:
	    b.pack(side=TOP)
	if not self.allowzero and self.btns == []:
	    b['relief'] = SUNKEN
	    self.selected_btn = [b]
	    self.value = [retval]
	    self.variable.set(`self.value`)
	self.btns.append(b)
	return b

    def btn_press(self, b):
	if b in self.selected_btn:
	    return
	if self.radio:
	    for btn in self.selected_btn:
		btn['relief'] = RAISED
	    self.selected_btn = [b]
	    self.value = [b.retval]
	else:
	    self.selected_btn.append(b)
	    self.value.append(b.retval)
	self.variable.set(self.value)
	b['relief'] = SUNKEN
	if self.user_proc:
	    self.user_proc(self.user_param, self.value)

# A paned window which consists of various horizontal / vertical panes.
# Panes are expected to be child widgets of this widget. The panes may be
# dynamically resized by dragging a 'sash' between two panes.
class PanedWindow(Frame):
    def __init__(self, master=None, orient=HORIZONTAL, cnf={}, **kw):
	if kw:
	    cnf = _cnfmerge((cnf, kw))
	else:
	    cnf = _cnfmerge(cnf)
	Frame.__init__(self, master, cnf)
	self.display_list = []
	self.orient = orient
	self.bind('<Configure>', self.compute_constraints)

    def __del__(self):
	for x in self.display_list:
	    x.__del__()
	self.destroy()

    def __getitem__(self, key):
	if key == 'orient':
	    return self.orient
	else:
	    Widget.__getitem__(self, key)

    def __setitem__(self, key, value):
	if key == 'orient':
	    if value in [HORIZONTAL, VERTICAL]:
		self.orient = value
		# Inform separators
		for obj in self.display_list:
		    if hasattr(obj, 'orient'):
			obj.set_orient(value)
		self.compute_constraints()
	else:
	    Widget.config(self, {key:value})

    # Find a pane in the display list given its name
    def _find_pane(self, name):
	for i in range(len(self.display_list)):
	    if self.display_list[i].name == name:
		return i
	return -1

    def pane(self, name):
	idx = self._find_pane(name)
	if idx == -1:
	    return None
	return self.display_list[idx]

    # Add a pane (and a separator if not the first pane)
    def add_pane(self, name, min_size=10, before=None, after=None):
	if self._find_pane(name) >= 0:
	    raise ValueError, 'Pane ' + name + ' already exists'
	idx = -1
	if before:
	    idx = self._find_pane(before)
	if idx == -1 and after:
	    idx = self._find_pane(after)
	pane = _Pane(self, name, min_size)
	if idx == -1:
	    if self.display_list == []:
		self.display_list.append(pane)
	    else:
		self.display_list.append(_PaneSeparator(self, self.orient))
		self.display_list.append(pane)
	    idx = 0
	else:
	    if idx == 0:
		self.display_list = [pane, _PaneSeparator(self, self.orient)] + self.display_list
	    else:
		if hasattr(display_list[idx - 1], 'orient'):
		    new_list = [pane, _PaneSeparator(self, self.orient)]
		else:
		    new_list = [_PaneSeparator(self, self.orient), pane]
		self.display_list = self.display_list[:idx] + new_list + self.display_list[idx:]

    # Remove a pane (and its associated separator)
    def remove_pane(self, name):
	idx = self._find_pane(name)
	if idx < 0:
	    raise ValueError, 'Pane ' + name + ' not found'
	self.display_list = self.display_list[:idx] + self.display_list[idx+1:]
	if idx >= len(self.display_list):
	    self.display_list = self.display_list[:-1]
	elif idx == 0:
	    self.display_list = self.display_list[1:]
	else:
	    self.display_list = self.display_list[:idx] + self.display_list[idx + 1:]

    # Compute x,y,width&height for all panes and separators
    def compute_constraints(self, event=None):
	if event:
	    width = event.width
	    height = event.height
	else:
	    width = self.winfo_width()
	    height = self.winfo_height()
	if width <=0:
	    width = 1
	if height <= 0:
	    height = 1
	tot_size = 0
	for i in range(len(self.display_list)):
	    tot_size = tot_size + self.display_list[i].min_size
	if self.orient == HORIZONTAL:
	    if tot_size > height:
		# too small to contain everything - increase canvas height
		self['height'] = tot_size
		expand = 1
	    else:
		# expand all panes by uniform scaling factor
		expand = float(height) / float(tot_size)
	else:
	    if tot_size > width:
		# too small to contain everything - increase canvas width
		self['width'] = tot_size
		expand = 1
	    else:
		expand = float(width) / float(tot_size)

	# Configure all the panes and separators
	dim = 0.0
	for obj in self.display_list[:-1]:
	    if hasattr(obj, 'orient'):
		inc = obj.min_size	# separator
	    else:
		inc = obj.min_size * expand # pane
	    if self.orient == HORIZONTAL:
		inc = float(inc) / float(height)
		obj.redimension(0, dim, 1.0, inc, width, height)
	    else:
		inc = float(inc) / float(width)
		obj.redimension(dim, 0, inc, 1.0, width, height)
	    dim = dim + inc
	# Last pane gets a little extra space because of rounding errors
	obj = self.display_list[len(self.display_list) - 1]
	if self.orient == HORIZONTAL:
	    obj.redimension(0, dim, 1.0, 1.0 - dim, width, height)
	else:
	    obj.redimension(dim, 0, 1.0 - dim, 1.0, width, height)

	self.draw()

    # Calculate the range of movement permissible for a separator (based
    # on its neighbouring panes minimum sizes). Return min, max values as
    # percentages of total width/height
    def calc_separator_range(self, sep_name):
	idx = self._find_pane(sep_name)
	obj = self.display_list[idx]
	prev = self.display_list[idx - 1]
	next = self.display_list[idx + 1]
	if self.orient == HORIZONTAL:
	    tot_height = self.winfo_height()
	    min = (prev.rely) * tot_height + prev.min_size
	    max = (next.rely + next.relheight) * tot_height - next.min_size
	    min = float(min) / float(tot_height)
	    max = float(max) / float(tot_height)
	    return (min, max)
	else:
	    tot_width = self.winfo_width()
	    min = (prev.relx) * tot_width + prev.min_size
	    max = (next.relx + next.relwidth) * tot_width - next.min_size
	    min = float(min) / float(tot_width)
	    max = float(max) / float(tot_width)
	    return (min, max)

    # Called during a drag of the separator - redraw neighbouring panes
    # x and y are relative (% of width & height)
    def resize_panes(self, sep_name, x, y, tot_width, tot_height):
	idx = self._find_pane(sep_name)
	pane1 = self.display_list[idx - 1]
	sep = self.display_list[idx]
	pane2 = self.display_list[idx + 1]
	if self.orient == HORIZONTAL:
	    p1h = y - pane1.rely
	    pane1.redimension(0, pane1.rely, 1.0, p1h, tot_width, tot_height)
	    p2y = y + sep.min_size / tot_height
	    p2h = pane2.relheight + pane2.rely - p2y
	    pane2.redimension(0, p2y, 1.0, p2h, tot_width, tot_height)
	else:
	    p1w = x - pane1.relx
	    pane1.redimension(pane1.relx, 0, p1w, 1.0, tot_width, tot_height)
	    p2x = x + sep.min_size / tot_width
	    p2w = pane2.relwidth + pane2.relx - p2x
	    pane2.redimension(p2x, 0, p2w, 1.0, tot_width, tot_height)
	self.draw()

    # Draw everything
    def draw(self):
	for x in self.display_list:
	    x.draw()

# A frame with a label. Actually implemented as a canvas. The frame is a
# subwidget. We need the canvas in order to draw the neat line around the
# frame.
class LabelledFrame(Frame):
    def __init__(self, master, text='', position=NW, cnf={}, **kw):
	if kw:
	    cnf = _cnfmerge((cnf, kw))
	else:
	    cnf = _cnfmerge(cnf)
	Frame.__init__(self, master, cnf)
	if position not in [NW, N, NE, SW, S, SE]:
	    position = NW
	self._label_pos = position

	# Create canvas items - dimensions will be overridden later
	self.label = Label(self, text=text)
	self.outer = Frame(self, height=6)
	self.frame = Frame(self, relief=GROOVE, bd=3)
	self._filler = Frame(self.frame, height=10)

	self.layout()

	self.bind('<Configure>', self.draw)

    def __getitem__(self, key):
	if key == 'position':
	    return self._label_pos
	elif key == 'text':
	    return self.label['text']
	else:
	    Widget.__getitem__(self, key)

    def __setitem__(self, key, value):
	if key == 'position':
	    if value in [N, NE, NW, S, SE, SW]:
		self._label_pos = value
		self.layout()
		self.draw()
	elif key == 'text':
	    self.label['text'] = value
	    self.draw()
	else:
	    Widget.config(self, {key:value})

    def layout(self):
	if self._label_pos[0] == N:
	    self._filler.pack(side=TOP, expand=0, fill=X)
	    self.outer.pack(side=TOP, expand=0, fill=X)
	    self.frame.pack(side=TOP, expand=1, fill=BOTH, padx=3, pady=3)
	    if self._label_pos == NW:
		self.label.place({'in':self.outer}, x=10, y=1, anchor=NW)
	    elif self._label_pos == N:
		self.label.place({'in':self.outer}, relx=0.5, y=1, anchor=N)
	    else:
		self.label.place({'in':self.outer}, relx=0.95, y=1, anchor=NE)
	else:
	    self._filler.pack(side=BOTTOM, expand=0, fill=X)
	    self.outer.pack(side=BOTTOM, expand=0, fill=X)
	    self.frame.pack(side=BOTTOM, expand=1, fill=BOTH, padx=3, pady=3)
	    if self._label_pos == SW:
		self.label.place({'in':self.outer}, x=10, rely=1.0, anchor=SW)
	    elif self._label_pos == S:
		self.label.place({'in':self.outer}, relx=0.5, rely=1.0, anchor=S)
	    else:
		self.label.place({'in':self.outer}, relx=0.95, rely=1.0, anchor=SE)

    def draw(self, event=None):
	self.label.tkraise()

#########################################################################
###                                                                   ###
### New combination widgets which accept restricted input types ---   ###
###           LabelledEntry, NumericEntry, DateEntry etc.             ###
###                                                                   ###
#########################################################################

# Entry widget + Label widget contained in a frame. If auto_select is
# non-zero, the contents of the entry widget are selected every time it gets
# the focus. command is called with cmd_param and the entry field value
# whenever the field is losing focus or <cr> is pressed.
# When the field is initially set, it contains text in black. Whenever it
# is modified by typing rather than programmatically, the field changes
# colour to blue. This is useful when we have a mix of computed and
# typed in fields on a form -- you can see at a glance which fields you
# have modified.
class LabelledEntry(Frame):
    def __init__(self, master=None, label=None, variable=None, auto_select=1,
		 command=None, cmd_param=None, normal_colour='black',
		 modified_colour='blue', labelwidth=None, entrywidth=None,
		 labelside=LEFT, anchor=NE, cnf={}, **kw):
	if kw:
	    cnf = _cnfmerge((cnf, kw))
	else:
	    cnf = _cnfmerge(cnf)
	Frame.__init__(self, master, cnf)

	if label:
	    self.label = Label(self, text=label, anchor=anchor)
	else:
	    self.label = Label(self, text='', anchor=anchor)
	self.labelside = labelside
	if labelwidth:
	    self.label['width'] = labelwidth
	self.label.pack(side=labelside)
	self.entry = Entry(self, relief=SUNKEN, border=2)
	if entrywidth:
	    self.entry['width'] = entrywidth
	self.entry.pack(side=LEFT, fill=X, expand=1)

	if variable:
	    self.value = variable
	else:
	    self.value = StringVar()
	self.old_value = ''
	self.entry['textvariable'] = self.value
	self.entry.bind('<Control-u>', self.clear)
	self.entry.bind('<FocusOut>', self.validate)
	self.entry.bind('<Return>', self.validate)
	if auto_select:
	    self.entry.bind('<FocusIn>', self._select_all)
	self.auto_select = auto_select
	self.entry.bind('<Control-space>', self._select_all)
	self.user_proc = command
	self.user_param = cmd_param
	self.norm_col = normal_colour
	self.mod_col = modified_colour
	self.modified = 0

    def __setitem__(self, key, value):
	if key in ['label', 'variable', 'auto_select', 'command', 'cmd_param',
		   'normal_colour', 'modified_colour', 'labelwidth',
		   'entrywidth', 'labelside', 'anchor']:
	    if key == 'label':
		self.label['text'] = value
	    elif key == 'variable':
		if value is None:
		    self.value = StringVar()
		else:
		    self.value = value
	    elif key == 'auto_select':
		if value:
		    self.entry.bind('<FocusIn>', self._select_all)
		else:
		    self.entry.bind('<FocusIn>', None)
		self.auto_select = value
	    elif key == 'command':
		self.user_proc = value
	    elif key == 'cmd_param':
		self.user_param = value
	    elif key == 'normal_colour':
		self.norm_col = value
	    elif key == 'modified_colour':
		self.mod_col = value
	    elif key == 'labelwidth':
		self.label['width'] = value
	    elif key == 'entrywidth':
		self.entry['width'] = value
	    elif key == 'labelside':
		self.labelside = value
		self.label.pack(side=value)
		self.entry.pack(side=LEFT, fill=X, expand=1)
	    elif key == 'anchor':
		self.label['anchor'] = value
	else:
	    Widget.config(self, {key: value})

    def __getitem__(self, key):
	if key in ['label', 'variable', 'auto_select', 'command', 'cmd_param',
		   'normal_colour', 'modified_colour', 'labelwidth',
		   'entrywidth', 'labelside', 'anchor']:
	    if key == 'label':
		return self.label['text']
	    elif key == 'variable':
		return self.variable
	    elif key == 'auto_select':
		return self.auto_select
	    elif key == 'command':
		return self.user_proc
	    elif key == 'cmd_param':
		return self.user_param
	    elif key == 'normal_colour':
		return self.norm_col
	    elif key == 'modified_colour':
		return self.mod_col
	    elif key == 'labelwidth':
		return self.label['width']
	    elif key == 'entrywidth':
		return self.entry['width']
	    elif key == 'labelside':
		return self.labelside
	    elif key == 'anchor':
		return self.label['anchor']
	else:
	    Widget.__getitem__(self, key)

    def clear(self, event=None):
	self.value.set('')
	self.modified = 0

    def revert(self):
	self.value.set(self.old_value)

    def get(self):
	return self.value.get()

    def set(self, text):
	self.old_value = text
	self.value.set(text)
	self.modified = 0
	self.entry['fg'] = self.norm_col

    def disable(self):
	self.entry['state'] = DISABLED

    def enable(self):
	self.entry['state'] = NORMAL

    def validate(self, event = None):
	new_value = self.value.get()
	if self.old_value != new_value:
	    if self.user_proc:
		stat = self.user_proc(self, self.user_param, self.get())
		if not stat:
		    self.entry.bell()
		    self.revert()
		    return
		elif stat == -1:
		    return
	    new_value = self.value.get() # in case user_proc called self.set()
	    self.old_value = new_value
	    self.modified = 1
	    self.entry['fg'] = self.mod_col
	return

    def _select_all(self, event = None):
	self.entry.select_range(0, END)

# Labelled entry field which only accepts numbers (including floats)
class NumericEntry(LabelledEntry):
    def __init__(self, master=None, label=None, variable=None, decimals=-1,
		 abbreviate=0, auto_select=1, auto_format=1, command=None,
		 cmd_param=None, normal_colour='black', modified_colour='blue',
		 labelwidth=None, entrywidth=None, labelside=LEFT, anchor=NE,
		 cnf={}, **kw):
	if kw:
	    cnf = _cnfmerge((cnf, kw))
	else:
	    cnf = _cnfmerge(cnf)
	LabelledEntry.__init__(self, master=master, label=label,
			       variable=variable, auto_select=auto_select,
			       command=command, cmd_param=cmd_param,
			       normal_colour=normal_colour,
			       modified_colour=modified_colour,
			       labelwidth=labelwidth, entrywidth=entrywidth,
			       labelside=labelside, anchor=anchor, cnf=cnf)
	self.decimals = decimals
	self.abbreviate = abbreviate
	self.auto_format = auto_format

    def __setitem__(self, key, value):
	if key in ['decimals', 'auto_format', 'abbreviate']:
	    if key == 'decimals':
		self.decimals = value
	    elif key == 'auto_format':
		self.auto_format = value
	    elif key == 'abbreviate':
		self.abbreviate = value
	else:
	    LabelledEntry.__setitem__(self, key, value)

    def __getitem__(self, key):
	if key in ['decimals', 'auto_format', 'abbreviate']:
	    if key == 'decimals':
		return self.decimals
	    elif key == 'auto_format':
		return self.auto_format
	    elif key == 'abbreviate':
		return self.abbreviate
	else:
	    LabelledEntry.__getitem__(self, key)

    def get(self):
	try:
	    val = Str2Amt(self.value.get())
	except ValueError:
	    val = None
	return val

    def validate(self, event = None):
	val = self.get()
	if val == None:
	    self.entry.bell()
	    self.revert()
	    return
	if self.auto_format:
	    self.value.set(Amt2Str(val, decimals=self.decimals,
				   use_abbrevs=self.abbreviate))
	new_value = self.value.get()
	if self.old_value != new_value:
	    if self.user_proc:
		stat = self.user_proc(self, self.user_param, self.get())
		if not stat:
		    self.entry.bell()
		    self.revert()
		    return
		elif stat == -1:
		    return
	    new_value = self.value.get() # in case user_proc called self.set()
	    self.old_value = new_value
	    self.modified = 1
	    self.entry['fg'] = self.mod_col
	return


# Entry to accept only valid dates. Also accepts strings of the form '1m', '2y'
# etc. in which case the date in the field is incremented by the appropriate
# number
class DateEntry(LabelledEntry):
    def __init__(self, master=None, label=None, variable=None,
		 auto_select=1, auto_format=1, command=None, cmd_param=None,
		 normal_colour='black', modified_colour='blue',
		 labelwidth=None, entrywidth=None, labelside=LEFT, anchor=NE,
		 cnf={}, **kw):
	if kw:
	    cnf = _cnfmerge((cnf, kw))
	else:
	    cnf = _cnfmerge(cnf)
	LabelledEntry.__init__(self, master=master, label=label,
			       variable=variable, auto_select=auto_select,
			       command=command, cmd_param=cmd_param,
			       normal_colour=normal_colour,
			       modified_colour=modified_colour,
			       labelwidth=labelwidth, entrywidth=entrywidth,
			       labelside=labelside, anchor=anchor, cnf=cnf)
	self.entry['width'] = 22
	self.auto_format = auto_format

    def __setitem__(self, key, value):
	if key == 'auto_format':
	    self.auto_format = value
	else:
	    LabelledEntry.__setitem__(self, key, value)

    def __getitem__(self, key):
	if key == 'auto_format':
	    return self.auto_format
	else:
	    LabelledEntry.__getitem__(self, key)

    def get(self):
	str = self.value.get()
	try:
	    val = DateHols.Date(str)
	except ValueError:
	    try:
		val = DateHols.YMDString(str)
		val = DateHols.Date(self.old_value) + str
	    except ValueError:
		val = None
	return val

    def validate(self, event = None):
	val = self.get()
	if val == None:
	    self.entry.bell()
	    self.revert()
	    return
	if self.auto_format:
	    if DateHols.IsDate(val):
		self.value.set(str(val))
	new_value = self.value.get()
	if self.old_value != new_value:
	    if self.user_proc:
		stat = self.user_proc(self, self.user_param, self.get())
		if not stat:
		    self.entry.bell()
		    self.revert()
		    return
		elif stat == -1:
		    return
	    new_value = self.value.get() # in case user_proc called self.set()
	    self.old_value = new_value
	    self.modified = 1
	    self.entry['fg'] = self.mod_col
	return

# Entry with a list of predefined entries which drop down if needed. The field
# may be constrained to accept only values from the list.
class ComboBox(LabelledEntry):
    def __init__(self, master=None, label=None, variable=None,
		 auto_select=1, opt_list=None, init_val=None, constrained=1,
		 command=None, cmd_param=None, normal_colour='black',
		 modified_colour='blue', labelwidth=None, entrywidth=None,
		 labelside=LEFT, anchor=NE, cnf={}, **kw):
	global _bitmap_images
	if kw:
	    cnf = _cnfmerge((cnf, kw))
	else:
	    cnf = _cnfmerge(cnf)
	LabelledEntry.__init__(self, master=master, label=label,
			       variable=variable, auto_select=auto_select,
			       command=command, cmd_param=cmd_param,
			       normal_colour=normal_colour,
			       modified_colour=modified_colour,
			       labelwidth=labelwidth, entrywidth=entrywidth,
			       labelside=labelside, anchor=anchor, cnf=cnf)

	self.constrained = constrained
	self.opt_list = self._ensure_list(opt_list)
	self.entry['width'] = self._list2strs() + 2

	self.list = PopupList(callback=self.popup_set, init_vals=self.opt_list,
			      ref_widget=self.entry)
	if not _bitmap_images.down_arrow:
	    _bitmap_images.load_images()
	self.help_btn = Button(self, image=_bitmap_images.down_arrow)
	self.help_btn['takefocus'] = 0
	self.help_btn.bind('<ButtonPress-1>', self.list.show_hide)
	self.popped_up = 0
	self.help_btn.pack(side=LEFT, expand=0)

	if type(init_val) != StringType:
	    init_val = `init_val`
	if self.constrained and init_val not in self.opt_list:
	    if len(self.opt_list) > 0:
		init_val = self.opt_list[0]
	self.value.set(init_val)
	self.old_value = init_val

    def _list2strs(self):
	maxlen = 1
	for i in range(len(self.opt_list)):
	    if type(self.opt_list[i]) != StringType:
		self.opt_list[i] = `self.opt_list[i]`
	    if len(self.opt_list[i]) > maxlen:
		maxlen = len(self.opt_list[i])
	return maxlen

    def _ensure_list(self, opt_list):
	if type(opt_list) == ListType:
	    opt_list = opt_list
	elif type(opt_list) == StringType:
	    opt_list = string.splitfields(opt_list, ',')
	elif type(opt_list) == TupleType:
	    opt_list = map(None, opt_list)
	else:
	    opt_list = []
	return opt_list

    def __setitem__(self, key, value):
	if key in ['opt_list', 'constrained']:
	    if key == 'constrained':
		self.constrained = value
	    elif key == 'opt_list':
		self.opt_list = self._ensure_list(value)
		self.entry['width'] = self._list2strs()
		self.list.listbox.delete(0, END)
		for opt in self.opt_list:
		    self.list.listbox.insert(END, opt)
	else:
	    LabelledEntry.__setitem__(self, key, value)

    def __getitem__(self, key):
	if key in ['opt_list', 'constrained']:
	    if key == 'constrained':
		return self.constrained
	    elif key == 'opt_list':
		return self.opt_list
	else:
	    LabelledEntry.__getitem__(self, key)

    def add(self, string):
	if string not in self.opt_list:
	    self.list.add(string)
	    self.opt_list.append(string)
	    self.entry['width'] = self._list2strs() + 2

    def set(self, val):
	if self.constrained and val not in self.opt_list:
	    val = self.opt_list[0]
	self.value.set(val)
	self.old_value = val

    def popup_set(self, val):
	self.value.set(val)
	self.validate()

    def get(self):
	s = self.value.get()
	if not self.constrained or s in self.opt_list:
	    return s
	# Constrained match - ignore case
	s = string.lower(s)
	matched = 0
	for x in self.opt_list:
	    lcase_option = string.lower(x)
	    if s == lcase_option:
		return x
	    elif string.find(lcase_option, s) == 0:
		matched = matched + 1
		ret_str = x
	if matched == 1:		# unique match
	    self.value.set(ret_str)
	    return ret_str
	return None

    def validate(self, event = None):
	val = self.get()
	if val == None:
	    self.entry.bell()
	    self.revert()
	    return
	new_value = self.value.get()
	if self.old_value != new_value:
	    if self.user_proc:
		stat = self.user_proc(self, self.user_param, self.get())
		if not stat:
		    self.entry.bell()
		    self.revert()
		    return
		elif stat == -1:
		    return
	    new_value = self.value.get() # in case user_proc called self.set()
	    self.old_value = new_value
	    self.modified = 1
	    self.entry['fg'] = self.mod_col
	return

# Fancier version of a Numeric field. Increment & decrement arrows at
# the right of the Entry widget change the numeric value in user defined
# chunks.
class NumericNudgeEntry(NumericEntry):
    def __init__(self, master=None, label=None, variable=None, decimals=-1,
		 abbreviate=0, auto_select=1, auto_format=1, command=None,
		 cmd_param=None, normal_colour='black', modified_colour='blue',
		 labelwidth=None, entrywidth=None, nudge=1, cnf={}, **kw):
	global _bitmap_images
	if kw:
	    cnf = _cnfmerge((cnf, kw))
	else:
	    cnf = _cnfmerge(cnf)
	NumericEntry.__init__(self, master=master, label=label,
			      variable=variable, decimals=decimals,
			      abbreviate=abbreviate, auto_select=auto_select,
			      auto_format=auto_format, command=command,
			      cmd_param=cmd_param, normal_colour=normal_colour,
			      modified_colour=modified_colour,
			      labelwidth=labelwidth, entrywidth=entrywidth,
			      cnf=cnf)
	self.nudge = nudge
	nudge_frame = Frame(self)
	if not _bitmap_images.incr_arrow:
	    _bitmap_images.load_images()
	btn_up = Button(nudge_frame, image=_bitmap_images.incr_arrow,
			command=self.increment)
	btn_down = Button(nudge_frame, image=_bitmap_images.decr_arrow,
			  command=self.decrement)
	btn_up.pack(side=TOP)
	btn_down.pack(side=BOTTOM)
	nudge_frame.pack(side=RIGHT)

    def __setitem__(self, key, value):
	if key == 'nudge':
	    self.nudge = value
	else:
	    NumericEntry.__setitem__(self, key, value)

    def __getitem__(self, key, value):
	if key == 'nudge':
	    return self.nudge
	else:
	    NumericEntry.__getitem__(self, key)

    def post_result(self, val):
	if self.auto_format:
	    self.value.set(Amt2Str(val, decimals=self.decimals,
				   use_abbrevs=self.abbreviate))
	else:
	    self.value.set(val)

    def increment(self):
	val = self.get()
	if val is None:
	    val = 0
	val = val + self.nudge
	self.post_result(val)
       	self.validate()

    def decrement(self):
	val = self.get()
	if val is None:
	    val = 0
	val = val - self.nudge
	self.post_result(val)
	self.validate()

# Fancier version of a Date field. Increment & decrement arrows at
# the right of the Entry widget change the date by user defined number of
# days.
class DateNudgeEntry(DateEntry):
    def __init__(self, master=None, label=None, variable=None,
		 auto_select=1, auto_format=1, command=None, cmd_param=None,
		 normal_colour='black', modified_colour='blue',
		 labelwidth=None, entrywidth=None, nudge=1, cnf={}, **kw):
	global _bitmap_images
	if kw:
	    cnf = _cnfmerge((cnf, kw))
	else:
	    cnf = _cnfmerge(cnf)
	DateEntry.__init__(self, master=master, label=label, variable=variable,
			   auto_select=auto_select, auto_format=auto_format,
			   command=command, cmd_param=cmd_param,
			   normal_colour=normal_colour,
			   modified_colour=modified_colour,
			   labelwidth=labelwidth, entrywidth=entrywidth,
			   cnf=cnf)
	if type(nudge) != IntType:
	    nudge = 1
	self.nudge = nudge
	nudge_frame = Frame(self)
	if not _bitmap_images.incr_arrow:
	    _bitmap_images.load_images()
	btn_up = Button(nudge_frame, image=_bitmap_images.incr_arrow,
			command=self.increment)
	btn_down = Button(nudge_frame, image=_bitmap_images.decr_arrow,
			  command=self.decrement)
	btn_up.pack(side=TOP)
	btn_down.pack(side=BOTTOM)
	nudge_frame.pack(side=RIGHT)

    def __setitem__(self, key, value):
	if key == 'nudge':
	    self.nudge = value
	else:
	    DateEntry.__setitem__(self, key, value)

    def __getitem__(self, key, value):
	if key == 'nudge':
	    return self.nudge
	else:
	    DateEntry.__getitem__(self, key)

    def increment(self):
	val = self.get()
	if val is None:
	    val = DateHols.Date()
	val = val + self.nudge
	self.value.set(str(val))
	self.validate()

    def decrement(self):
	val = self.get()
	if val is None:
	    val = DateHols.Date()
	val = val - self.nudge
	self.value.set(str(val))
	self.validate()

#### -------------------------------------------------------------------------
#### Everything below this point is internal to TkXtra and is not meant for
#### public consumption.
#### -------------------------------------------------------------------------

# Bitmap used for the ComboBox widget
_DOWN_ARROW_BITMAP = """
#define down_arrow_width 11
#define down_arrow_height 14
static char down_arrow_bits[] = {
    0x00, 0x00, 0x70, 0x00, 0x70, 0x00, 0x70, 0x00, 0x70, 0x00, 0x70, 0x00,
    0xfe, 0x03, 0xfc, 0x01, 0xf8, 0x00, 0x70, 0x00, 0x20, 0x00, 0x00, 0x00,
    0xfe, 0x03, 0xfe, 0x03
};
"""

# Bitmaps used for the *NudgeEntry compound widgets
_INCR_ARROW_BITMAP = """
#define incr_width 7
#define incr_height 4
static char incr_bits[] = {
    0x08, 0x1c, 0x3e, 0x7f
};
"""

_DECR_ARROW_BITMAP = """
#define decr_width 7
#define decr_height 4
static char decr_bits[] = {
    0x7f, 0x3e, 0x1c, 0x08
};
"""

class _TkXtraBitmaps:
    def __init__(self):
	self.down_arrow = None
	self.incr_arrow = None
	self.decr_arrow = None

    def load_images(self):
	# Below can't be done unless Tk has been initialized
	self.down_arrow = Image('bitmap', data=_DOWN_ARROW_BITMAP)
	self.incr_arrow = Image('bitmap', data=_INCR_ARROW_BITMAP)
	self.decr_arrow = Image('bitmap', data=_DECR_ARROW_BITMAP)

_bitmap_images = _TkXtraBitmaps()



#******************************
#***** Paned Window Stuff *****
#******************************

# A single pane of a PanedWindow
class _Pane(Frame):
    def __init__(self, master, name, min_size):
	Frame.__init__(self, master, relief=RAISED, bd=2)
	self.name = name
	if min_size < 1:
	    min_size = 1
	self.min_size = min_size
	self.relx = self.rely = 0
	self.relwidth = self.relheight = 1

    def redimension(self, x, y, w, h, tot_width, tot_height):
	# Called by parent when doing geometry management
	self.relx = x
	self.rely = y
	self.relwidth = w
	self.relheight = h

    def draw(self):
	self.place(relx=self.relx, rely=self.rely, relwidth=self.relwidth,
		   relheight=self.relheight)

# The pane sash - assumes it is parented by a PanedWindow
class _PaneSeparator:
    def __init__(self, master, orient=HORIZONTAL):
	self.name = _generate_name()
	self.min_size = 0
	self.min_val = 0		# set up during the start of a drag
	self.max_val = 1
	self.sep = Frame(master, width=3, height=3, bd=2, relief=RIDGE)
	self.grip = Frame(master, width=10, height=10, bd=2, relief=RAISED,
			  cursor='crosshair')
	self.orient = orient
	self.relx = self.rely = 0

	self.master = master
	self.x_root = self.y_root = 0
	self.width = self.height = 1	# dimensions of PARENT

	# Mouse events to give feedback during resize
	self.grip.bind('<1>', self._drag_start)
	self.grip.bind('<ButtonRelease-1>', self._drag_end)
	self.grip.bind('<B1-Motion>', self._drag_feedback)

    def __del__(self):
	self.master = None
	self.grip.destroy()
	self.sep.destroy()

    def redimension(self, x, y, w, h, tot_width, tot_height):
	# Called by master parent when doing geometry management
	if self.orient == HORIZONTAL:
	    self.relx = 0.95
	    self.rely = y
	else:
	    self.relx = x
	    self.rely = 0.95
	self.width = tot_width
	self.height = tot_height

    # Dynamically change our orientation
    def set_orient(self, value):
	if value == self.orient:
	    return
	self.orient = value
	# Just changing placement leads to infinite loops in Tk. So we
	# just recreate the separator widget
	self.sep.destroy()
	self.sep = Frame(self.master, width=3, height=3, bd=2, relief=RIDGE)
	if value == HORIZONTAL:
	    self.sep.place(relx=0, rely=self.rely, height=3, relwidth=1.0,
			   anchor=W)
	else:
	    self.sep.place(relx=self.relx, rely=0, relheight=1.0, width=3,
			   anchor=N)
	self.sep.lower()

    # Start a drag of the sash
    def _drag_start(self, event=None):
	self.sep.tkraise()
	self.grip.tkraise()
	# We need the x_root, y_root because the x & y in the event seem
	# to be highly unreliable (bug in Tk ??)
	self.x_root = self.master.winfo_rootx()
	self.y_root = self.master.winfo_rooty()
	self.width = self.master.winfo_width()
	self.height = self.master.winfo_height()
	self.min_val, self.max_val = self.master.calc_separator_range(self.name)

    # Allow a drag within the bounds specified by min_val and max_val
    def _drag_feedback(self, event=None):
	# We need the x_root, y_root because the x & y in the event seem
	# to be highly unreliable (bug in Tk ??)
	if self.orient == HORIZONTAL:
	    y = float(event.y_root - self.y_root) / float(self.height)
	    if y >= self.min_val and y <= self.max_val:
		self.rely = y
	else:
	    x = float(event.x_root - self.x_root) / float(self.width)
	    if x >= self.min_val and x <= self.max_val:
		self.relx = x
	self.draw()

    # Resize the two bordering panes
    def _drag_end(self, event=None):
	self.sep.lower()
	self.master.resize_panes(self.name, self.relx, self.rely,
				 self.width, self.height)

    def tkraise(self):
	self.grip.tkraise()

    # Draw grip and separator
    def draw(self):
	if self.orient == HORIZONTAL:
	    self.grip.place(relx=self.relx, rely=self.rely, anchor=CENTER)
	    self.sep.place(relx=0, rely=self.rely, height=3, relwidth=1.0,
			   anchor=W)
	else:
	    self.grip.place(relx=self.relx, rely=self.rely, anchor=CENTER)
	    self.sep.place(relx=self.relx, rely=0, relheight=1.0, width=3,
			   anchor=N)

# Generator for paned window separator names
_sep_num = 0
def _generate_name():
    global _sep_num
    _sep_num = _sep_num + 1
    return '__sep_' + `_sep_num`

### End of TkXtra.py
